from __future__ import print_function

import itertools
import pandas as pd
import numpy as np
import os
import logging
import networkx as nx
import math
from tqdm import tqdm

def buildNetFile(sampdirs, netfile, cutoff, auxDir, writecomponents=False):
    logger = logging.getLogger("grouper")

    sep = os.path.sep
    sffiles = [sep.join([sd, 'quant.sf']) for sd in sampdirs]
    eqfiles = [sep.join([sd, auxDir, '/eq_classes.txt']) for sd in sampdirs]

    tnames = []
    weightDict = {}
    diagCounts = None
    sumCounts = None
    ambigCounts = None
    firstSamp = True
    numSamp = 0
    tot = 0
    eqClasses = {}
    for sffile, eqfile in itertools.izip(sffiles, eqfiles):
        quant = pd.read_table(sffile)
        quant.set_index('Name', inplace=True)

        with open(eqfile) as ifile:
            numSamp += 1
            numTran = int(ifile.readline().rstrip())
            numEq = int(ifile.readline().rstrip())
            logging.info("quant file: {}; eq file: {}; # tran = {}; # eq = {}".format(sffile, eqfile, numTran, numEq))
            if firstSamp:
                for i in xrange(numTran):
                    tnames.append(ifile.readline().rstrip())
                diagCounts = np.zeros(len(tnames))
                sumCounts = np.zeros(len(tnames))
                ambigCounts = np.zeros(len(tnames))
            else:
                for i in xrange(numTran):
                    ifile.readline()

            # for easy access to quantities of interest
            tpm = quant.loc[tnames, 'TPM'].values
            estCount = quant.loc[tnames, 'NumReads'].values
            efflens = quant.loc[tnames, 'EffectiveLength'].values
            epsilon =  np.finfo(float).eps
            sumCounts = np.maximum(sumCounts, estCount)

            for i in xrange(numEq):
                toks = map(int, ifile.readline().rstrip().split('\t'))
                nt = toks[0]
                tids = tuple(toks[1:-1])
                count = toks[-1]
                if tids in eqClasses:
                    eqClasses[tids] += count
                else:
                    eqClasses[tids] = count

                # Add the contribution to the graph
                denom = sum([tpm[t] for t in tids])
                for t1, t2 in itertools.combinations(tids,2):
                    w = count
                    key = (t1, t2)
                    if key in weightDict:
                        weightDict[key] += w
                    else:
                        weightDict[key] = w
                for t in tids:
                    diagCounts[t] += count * (tpm[t] / denom)
                    ambigCounts[t] += count
            firstSamp = False

    lens = quant.loc[tnames, 'Length'].values

    maxWeight = 0.0
    minWeight = 0.0
    prior = 0.1
    edgesToRemove = []

    ##
    #  Go through the weightMap and remove any edges that
    #  have endpoints with too few mapping reads
    ##
    for k,v in weightDict.iteritems():
        c0, c1 = diagCounts[k[0]], diagCounts[k[1]]
        a0, a1 = ambigCounts[k[0]], ambigCounts[k[1]]
        if a0 + a1 > epsilon and a0 > cutoff and a1 > cutoff:
            w = (v+prior) / min((a0+prior), (a1+prior))
            if w > minWeight:
                weightDict[k] = w
                if w > maxWeight:
                    maxWeight = w
            else:
                edgesToRemove.append(k)
        else:
            edgesToRemove.append(k)

    # Actually delete those edges
    for e in edgesToRemove:
        del weightDict[e]

    def nearEnd(tup):
        txp = tup[0]
        pos = tup[1]
        moverhang = 10
        ml = 100
        if pos < -moverhang or pos > lens[txp] + moverhang:
            return False
        elif pos <= ml or pos >= lens[txp] - ml:
            return True
        else:
            return False

    tnamesFilt = []
    relabel = {}
    for i in xrange(len(estCount)):
        if (ambigCounts[i] > cutoff):
            relabel[i] = len(tnamesFilt)
            tnamesFilt.append(tnames[i])
            weightDict[(i, i)] = 1.1

    G = nx.Graph() if writecomponents else None
    with open(netfile, 'w') as ofile:
        writeEdgeList(weightDict, tnames, ofile, G)

    if G is not None:
        clustFile = netfile.split('.net')[0] + '.clust'
        print("Writing connected components as clusters to {}".format(clustFile))
        with open(clustFile, 'w') as ofile:
            cc = nx.connected_component_subgraphs(G)
            for c in cc:
                ofile.write('{}\n'.format('\t'.join(c.nodes())))

def writeEdgeList(weightDict, tnames, ofile, G):
    useGraph = G is not None
    for k,v in weightDict.items():
        ofile.write("{}\t{}\t{}\n".format(tnames[k[0]], tnames[k[1]], v))
        if useGraph:
            G.add_edge(tnames[k[0]], tnames[k[1]])


def writePajek(weightDict, tnames, relabel, ofile):
    with open(netfile, 'w') as ofile:
        ofile.write("*Vertices\t{}\n".format(len(tnamesFilt)))
        for i, n in enumerate(tnamesFilt):
            ofile.write("{}\t\"{}\"\n".format(i, n))
        ofile.write("*Edges\n")
        print("There are {} edges\n".format(len(weightDict)))
        for k,v in weightDict.items():
            ofile.write("{}\t{}\t{}\n".format(relabel[k[0]], relabel[k[1]], v))

class EquivCollection(object):
    def __init__(self):
        self.tnames = []
        self.eqClasses = {}
        self.hasNames = False

    def setNames(self, names):
        self.tnames = names
        self.hasNames = True

    def add(self, tids, count):
        if tids in self.eqClasses:
            self.eqClasses[tids] += count
        else:
            self.eqClasses[tids] = count

def readEqClass(eqfile, eqCollection):
    with open(eqfile) as ifile:
        numTran = int(ifile.readline().rstrip())
        numEq = int(ifile.readline().rstrip())
        print("file: {}; # tran = {}; # eq = {}".format(eqfile, numTran, numEq))
        if not eqCollection.hasNames:
            tnames = []
            for i in range(numTran):
                tnames.append(ifile.readline().rstrip())
            eqCollection.setNames(tnames)
        else:
            for i in range(numTran):
                ifile.readline()

        for i in range(numEq):
            toks = list(map(int, ifile.readline().rstrip().split('\t')))
            nt = toks[0]
            tids = tuple(toks[1:-1])
            count = toks[-1]
            eqCollection.add(tids, count)

def getCountsFromEquiv(eqCollection):
    countDict = {}
    tn = eqCollection.tnames
    for tids, count in eqCollection.eqClasses.items():
        for t in tids:
            if tn[t] in countDict:
                countDict[tn[t]] += count
            else:
                countDict[tn[t]] = count
    # ensure no division by 0
    for t in eqCollection.tnames:
        if t in countDict:
            countDict[t] += 1.0
        else:
            countDict[t] = 1.0
    return countDict

def flattenClusters(infile, outfile):
    with open(outfile, 'w') as ofile:
        with open(infile) as ifile:
            for i,l in enumerate(ifile):
                toks = l.rstrip().split()
                cname = "cluster{}".format(i)
                for t in toks:
                    ofile.write("{}\t{}\n".format(cname, t))

def filterGraph(expDict, netfile, outfile, auxDir, mincut):
    logger = logging.getLogger("grouper")
    # Get just the set of condition names
    conditions = expDict.keys()
    logging.info("conditions = {}".format(conditions))

    eqClasses = {}
    for cond in conditions:
        print(expDict[cond])
        for sampNum, sampPath in expDict[cond].items():
            if cond not in eqClasses:
                eqClasses[cond] = EquivCollection()
            eqPath = os.path.sep.join([sampPath, auxDir, "/eq_classes.txt"])
            readEqClass(eqPath, eqClasses[cond])

    ambigCounts = {cond : getCountsFromEquiv(eqClasses[cond]) for cond in conditions}

    sailfish = {}
    for cond in conditions:
        sailfish[cond] = ambigCounts[cond]

    G = nx.Graph()
    with open(netfile) as f:
        data = pd.read_table(f, header=None)
        for i in range(len(data)):
            G.add_edge(data[0][i], data[1][i], capacity = float(data[2][i]))

    logging.info("Done reading files for filtering.")
    count = 0
    numTrimmed = 0
    numCut = 0
    cutset = set()
    with open(netfile) as f, open(outfile, 'w') as ofile:
        data = pd.read_table(f, header=None)
        #for i in tqdm(range(len(data))):
        for i in range(len(data)):
            count += 1
            print("\r{} edges checked".format(count), end="")
            #Alternative hypo
            x = data[0][i]
            y = data[1][i]
            non_null=0
            x_all=0
            y_all=0
            for cond in conditions:
                y_g = sailfish[cond][y]
                x_g = sailfish[cond][x]
                r = y_g / x_g
                non_null += (y_g * math.log(r*x_g)) - (r*x_g)
                non_null += (x_g * math.log(x_g)) - x_g
                x_all += x_g
                y_all += y_g
            #null hypothesis
            null = 0
            r_all = y_all / x_all
            for cond in conditions:
                y_g = sailfish[cond][y]
                x_g = sailfish[cond][x]
                mean_x = (x_g + y_g) / (1+r_all)
                null += (y_g * math.log(r_all * mean_x)) - (r_all * mean_x)
                null += (x_g * math.log(mean_x)) - mean_x
            D = 2*(non_null-null)

            if mincut:
                if D > 20 and x != y:
                    s = G.subgraph(nx.shortest_path(G,x))
                    value, partition =  nx.minimum_cut(s, x, y)
                    if value < 10:
                        reachable, non_reachable = partition
                        for e in s.edges_iter(data='capacity'):
                            if (e[0] in reachable and e[1] in non_reachable) or (e[0] in non_reachable and e[1] in reachable):
                                cutset.add((e[0], e[1]))
                        numCut += 1
                    numTrimmed += 1
            else:
                if D > 20 and x != y:
                    numTrimmed += 1
                    cutset.add((x, y))

        G.remove_edges_from(list(cutset))
        for e in G.edges(data = "capacity"):
            ofile.write(e[0] + "\t" + e[1] + "\t" + str(e[2]) + "\n")

    logging.info("Trimmed {} edges".format(numTrimmed))
    logging.info("Cut performed on {} edges".format(numCut))

def addOrphanLinks(sampdirs, auxDir, orphanFileName, cutoff, netFileIn, netFileOut):
    logger = logging.getLogger("grouper")

    sep = os.path.sep
    sffiles = [sep.join([sd, 'quant.sf']) for sd in sampdirs]
    eqfiles = [sep.join([sd, auxDir, '/eq_classes.txt']) for sd in sampdirs]

    tnames = []
    weightDict = {}
    diagCounts = None
    sumCounts = None
    ambigCounts = None
    firstSamp = True
    numSamp = 0
    eqClasses = {}
    for sffile, eqfile in itertools.izip(sffiles, eqfiles):
        quant = pd.read_table(sffile)
        quant.set_index('Name', inplace=True)

        with open(eqfile) as ifile:
            numSamp += 1
            numTran = int(ifile.readline().rstrip())
            numEq = int(ifile.readline().rstrip())
            if firstSamp:
                for i in range(numTran):
                    tnames.append(ifile.readline().rstrip())
                diagCounts = np.zeros(len(tnames))
                sumCounts = np.zeros(len(tnames))
                ambigCounts = np.zeros(len(tnames))
            else:
                for i in range(numTran):
                    ifile.readline()

            # for easy access to quantities of interest
            tpm = quant.loc[tnames, 'TPM'].values
            estCount = quant.loc[tnames, 'NumReads'].values
            efflens = quant.loc[tnames, 'EffectiveLength'].values
            epsilon =  np.finfo(float).eps
            sumCounts = np.maximum(sumCounts, estCount)

            for i in range(numEq):
                toks = map(int, ifile.readline().rstrip().split('\t'))
                nt = toks[0]
                tids = tuple(toks[1:-1])
                count = toks[-1]
                if tids in eqClasses:
                    eqClasses[tids] += count
                else:
                    eqClasses[tids] = count

                denom = sum([tpm[t] for t in tids])
                for t in tids:
                    diagCounts[t] += count * (tpm[t] / denom)
                    ambigCounts[t] += count

            firstSamp = False

    lens = quant.loc[tnames, 'Length'].values
    logging.info("Done reading files for adding orphans")
    # Considering Orphan reads
    def nearEnd(tup):
        txp = tup[0]
        pos = tup[1]
        moverhang = 10
        ml = 100
        if pos < -moverhang or pos > lens[txp] + moverhang:
            return False
        elif pos <= ml or pos >= lens[txp] - ml:
            return True
        else:
            return False

    vertices = set()
    with open(netFileIn, 'r') as ifile:
        for line in ifile:
            line = line.split()
            vertices.add(line[0])
            vertices.add(line[1])

    count = 0
    seenOrphan = {}
    orphanDict = {}
    orphanLinkFiles = [sep.join([sd, auxDir, orphanFileName]) for sd in sampdirs]
    haveLinkFiles = all(os.path.isfile(f) for f in orphanLinkFiles)
    numOrphanLinks = 0
    if haveLinkFiles:
        for olfile in orphanLinkFiles:
            for l in open(olfile):
                left, right = l.rstrip().split(':')
                lp = [map(int, i.split(',')) for i in left.rstrip('\t').split('\t')]
                rp = [map(int, i.split(',')) for i in right.split('\t')]
                lp = [t for t in filter(nearEnd, lp)]
                rp = [t for t in filter(nearEnd, rp)]
                #if len(lp) == 1 or len(rp) == 1:
                for a, b in itertools.product(lp, rp):
                    ltpm = tpm[a[0]] + 10 ** -11  # Laplacian Smoothing
                    rtpm = tpm[b[0]] + 10 ** -11
                    tpm_ratio = 1 - (abs(ltpm - rtpm) / (ltpm + rtpm))
                    read_dist = lens[a[0]] - a[1] + b[1]
                    if tpm_ratio >= .5: # and read_dist <= 300 and tpm[a[0]] > .5 and tpm[b[0]] > .5:
                        a = a[0]; b = b[0]
                        key = (a, b) if a < b else (b, a)
                        if ambigCounts[a] < cutoff or ambigCounts[b] < cutoff:
                            continue
                        c0, c1 = diagCounts[a], diagCounts[b]
                        a0, a1 = ambigCounts[a], ambigCounts[b]
                        if a not in seenOrphan:
                            seenOrphan[a] = set()
                        if b not in seenOrphan:
                            seenOrphan[b] = set()
                        seenOrphan[a].add(b)
                        seenOrphan[b].add(a)
                        if key not in orphanDict:
                            orphanDict[key] = 1.0 / min(a0, a1)
                        else:
                            orphanDict[key] += 1.0 / min(a0, a1)

    for key, value in orphanDict.iteritems():
        if len(seenOrphan[key[0]]) < 3 and len(seenOrphan[key[1]]) < 3:
            if key not in weightDict:
                numOrphanLinks += 1
                weightDict[key] = 1.0 / min(a0, a1)
            else:
                weightDict[key] += 1.0 / min(a0, a1)
                
    logging.info("Added {} orphan link edges".format(numOrphanLinks))

    with open(netFileIn, 'r') as ifile, open(netFileOut, 'w') as ofile:
        for line in ifile:
            ofile.write(line)
        for k,v in weightDict.items():
            ofile.write("{}\t{}\t{}\n".format(tnames[k[0]], tnames[k[1]], v))

    return weightDict;



from __future__ import print_function
import os
import pandas as pd
import math
import logging

class EquivCollection(object):
    def __init__(self):
        self.tnames = []
        self.eqClasses = {}
        self.hasNames = False

    def setNames(self, names):
        self.tnames = names
        self.hasNames = True

    def add(self, tids, count):
        if tids in self.eqClasses:
            self.eqClasses[tids] += count
        else:
            self.eqClasses[tids] = count

def readEqClass(eqfile, eqCollection):
    with open(eqfile) as ifile:
        numTran = int(ifile.readline().rstrip())
        numEq = int(ifile.readline().rstrip())
        if not eqCollection.hasNames:
            tnames = []
            for i in xrange(numTran):
                tnames.append(ifile.readline().rstrip())
            eqCollection.setNames(tnames)
        else:
            for i in xrange(numTran):
                ifile.readline()

        for i in xrange(numEq):
            toks = map(int, ifile.readline().rstrip().split('\t'))
            nt = toks[0]
            tids = tuple(toks[1:-1])
            count = toks[-1]
            eqCollection.add(tids, count)

def getCountsFromEquiv(eqCollection):
    countDict = {}
    tn = eqCollection.tnames
    for tids, count in eqCollection.eqClasses.iteritems():
        for t in tids:
            if tn[t] in countDict:
                countDict[tn[t]] += count
            else:
                countDict[tn[t]] = count
    # ensure no division by 0
    for t in eqCollection.tnames:
        if t in countDict:
            countDict[t] += 1.0
        else:
            countDict[t] = 1.0
    return countDict

def filter(expDict, netfile, ofile, auxDir):
    logger = logging.getLogger("grouper")
    # Get just the set of condition names
    conditions = expDict.keys()
    print("conditions = {}".format(conditions))

    eqClasses = {}
    for cond in conditions:
        print(expDict[cond])
        for sampNum, sampPath in expDict[cond].iteritems():
            if cond not in eqClasses:
                eqClasses[cond] = EquivCollection()
            eqPath = os.path.sep.join([sampPath, auxDir, "eq_classes.txt"])
            readEqClass(eqPath, eqClasses[cond])

    ambigCounts = {cond : getCountsFromEquiv(eqClasses[cond]) for cond in conditions}

    sailfish = {}
    for cond in conditions:
        sailfish[cond] = ambigCounts[cond]

    count = 0
    numTrimmed = 0
    with open(netfile) as f, open(ofile, 'w') as ofile:
        data = pd.read_table(f, header=None)
        for i in range(len(data)):
            count += 1
            print("\r{} done".format(count), end="")
            #Alternative hypo
            x = data[0][i]
            y = data[1][i]
            non_null=0
            x_all=0
            y_all=0
            for cond in conditions:
                y_g = sailfish[cond][y]
                x_g = sailfish[cond][x]
                r = y_g / x_g
                non_null += (y_g * math.log(r*x_g)) - (r*x_g)
                non_null += (x_g * math.log(x_g)) - x_g
                x_all += x_g
                y_all += y_g
            #null hypothesis
            null = 0
            r_all = y_all / x_all
            for cond in conditions:
                y_g = sailfish[cond][y]
                x_g = sailfish[cond][x]
                mean_x = (x_g + y_g) / (1+r_all)
                null += (y_g * math.log(r_all * mean_x)) - (r_all * mean_x)
                null += (x_g * math.log(mean_x)) - mean_x
            D = 2*(non_null-null)
            if D <= 20:
                ofile.write("{}\t{}\t{}\n".format(x, y, data[2][i]))
            else:
                numTrimmed += 1
    logging.info("Trimmed {} edges after label propagation.".format(numTrimmed))



import os
from subprocess import call
import pandas as pd
from collections import defaultdict
import logging

#run BLAST and return list of label files in order [TS->ASdb, AS->TSdb] 
def runBLAST(faFile1, faFile2, outdir, threads):
	logger = logging.getLogger("grouper")
	TSdb = os.path.sep.join([outdir, "TS", "db"])
	ASdb = os.path.sep.join([outdir, "AS", "db"])
	call(["makeblastdb", "-in", faFile1, "-dbtype", "nucl", "-out", TSdb])
	call(["makeblastdb", "-in", faFile2, "-dbtype", "nucl", "-out", ASdb])

	labelFile1 = os.path.sep.join([outdir, "TS.ASdb" ]) 
	labelFile2 = os.path.sep.join([outdir, "AS.TSdb" ]) 
	#run blastn for TS against AS
	call(["blastn", "-db", ASdb, "-query", faFile1, "-outfmt", "6", "-out", labelFile1, "-num_threads", str(threads)])
	call(["blastn", "-db", TSdb, "-query", faFile2, "-outfmt", "6", "-out", labelFile2, "-num_threads", str(threads)])
	
	logging.info("Done running BLAST using {} threads".format(threads))
	return [labelFile1, labelFile2]

#get the lengths of transcripts from one of the quant files
def getLengths(expDict):
    logger = logging.getLogger("grouper")
    lengths = {}
    condition = list(expDict)[0]
    quantFile = os.path.sep.join([expDict[condition][0], "quant.sf"])
    logging.info("Getting txp lengths from: {}".format(quantFile))

    lengths = pd.read_table(quantFile, names=["Name", "Length"], usecols=["Name", "Length"])
    lengths.set_index("Name", inplace=True)
    lengths = lengths.to_dict()['Length']

    return lengths

#read in the BLAST results and process them to get two way best match. This creates the seed file for Junto.
#Label files should be in order [TS->ASdb, AS->TSdb] 
def genFinalLabels(expDict, keys, labelFiles, finalLabelFile, outdir):
	lengths = getLengths(expDict)

	blastOut = open(keys["seed_file"], 'w')
	if (len(labelFiles)==1):
		with open(labelFiles[0].strip(), 'r') as f:
			data = pd.read_table(f, header=None, usecols=[0,1], names=['query', 'subject'])
			table1 = data.set_index("query").to_dict()['subject']
		try:
			for key,value in table1.items():
				probability=1.0
				blastOut.write(key + "\t" + value + "\t" +str(probability)+ "\n")
		except KeyError:
			pass
	else:
		labelFiles[0] = labelFiles[0].strip()
		labelFiles[1] = labelFiles[1].strip()
		outfile1 = outdir + "best" + os.path.basename(labelFiles[0])
		outfile2 = outdir + "best" + os.path.basename(labelFiles[1])
		command = "cat " + labelFiles[0]  + " |  sort -k1,1 -k12,12nr -k11,11n | sort -u -k1,1 --merge > " + outfile1
		call(command, shell=True)
		command = "cat " + labelFiles[1] + " |  sort -k2,2 -k12,12nr -k11,11n | sort -u -k2,2 --merge > " + outfile2
		call(command, shell=True)
		with open(outfile1, 'r') as f:
			table1 = {}
			lentable1 = {}
			for line in f:
				line = line.split()
				table1[line[0]] = line[1]
				lentable1[line[0]] = abs(float(line[6]) - float(line[7]))

		with open(outfile2, 'r') as f:
			table2 = {}
			lentable2 = {}
			for line in f:
				line = line.split()
				table2[line[1]] = line[0]
				lentable2[line[1]] = abs(float(line[8]) - float(line[9]))	

		call(["rm", outfile1])
		call(["rm", outfile2])

		try:
			for key,value in table1.items():
				if key in table2:
					alignLen = max(lentable1[key], lentable2[key])
					probability=min(1.0, (alignLen / float(lengths[key])))
					blastOut.write(key + "\t" + value + "\t" +str(probability)+ "\n")
		except KeyError:
				pass
	blastOut.close()
	call(["cp", keys["seed_file"], finalLabelFile])



import sys
import os
import argparse
import itertools
import statistics
import re
import csv
import fileinput
import numpy as np
from collections import defaultdict
from subprocess import call
import pandas as pd
from collections import Counter

alpha = 0.8 #ratio alloted to the probability value for labels
newEdgeProb = 0.9 #add a new edge if the shared label prob is greater than this
                    
def getMedWeight(graph, node1, node2):
	weights = []
	for (x, weight) in graph[node1]:
		if weight != 1.1:
			weights.append(weight)
	for (x, weight) in graph[node2]:
		if weight != 1.1:
			weights.append(weight)
	
	if not weights:
		return(0)
	else:
		return(statistics.median(weights))

def readLabels(keys, iterNum):
	#below we get labelToContigs = labels -> contigs that map to this label
	#and contigToLabels = contigs -> all labels
	#contigLabelsToProb = contig, labels -> prob
	labelToContigs = defaultdict(set)
	contigToLabels = defaultdict(set)
	contigLabelsToProb = {}
	with open(keys["seed_file"], 'r') as ifile:
		for line in ifile:
			data = (line.strip('\n')).split('\t')
			contigName = data[0]
			if (iterNum >= 1):
				data = data[3].split()
				curLabel = ""
				for i in range(1, len(data), 2):
					curLabel = data[i-1]
					if (curLabel != "" and curLabel != "__DUMMY__" and data[i].lower() != "nan" and curLabel != "IGNORE"):
						contigToLabels[contigName].add(curLabel)
						labelToContigs[curLabel].add(contigName)
						contigLabelsToProb[(contigName, curLabel)] = float(data[i])
			else:
				curLabel = data[1]
				contigToLabels[contigName].add(curLabel)
				labelToContigs[curLabel].add(contigName)
				contigLabelsToProb[(contigName, curLabel)] = float(data[i])
	return(labelToContigs, contigToLabels, contigLabelsToProb)

#get the sum of the probabilities that 2 nodes have the same label
#@profile
def getProb(contigToLabels, contigLabelsToProb, node1, node2):
	totalProbability = 0
	sharedLabels = contigToLabels[node1].intersection(contigToLabels[node2])
	for label in sharedLabels:
		totalProbability += (contigLabelsToProb[(node1, label)] * contigLabelsToProb[(node2, label)])
	return totalProbability

#@profile
def changeEdgeWeights(orgGraph, graph, contigToLabels, contigLabelsToProb, ofile):
	changesMade = 0
	noChange = 0
	weightCalc = 0
	selfLoops = 0

	for (node1, node2), weight in graph.iteritems():
		if node1 == node2:
			ofile.write(node1 + "\t" + node2 + "\t1.1\n")
			selfLoops += 1
		else:
			prob = getProb(contigToLabels, contigLabelsToProb, node1, node2)
			if prob > 0:
				for (x, w) in orgGraph[node1]:
					if x == node2:
						orgWeight = w
				for (x, w) in orgGraph[node2]:
					if x == node1:
						orgWeight = w
				newWeight = ((1 - alpha) * orgWeight) + (prob * alpha)
				changesMade += 1
			else:
				newWeight = weight
				noChange += 1
			weightCalc += newWeight
			ofile.write(node1 + "\t" + node2 + "\t" + str(newWeight) + "\n")
	weightCalc /= (changesMade+noChange)

	return (weightCalc, (changesMade + noChange + selfLoops))

#@profile
def addNewEdges(orgGraph, graph, contigToLabels, labelToContigs, contigLabelsToProb, ofile):
	changesMade = 0
	weightCalc = 0
	for label in labelToContigs:
		for node1, node2 in itertools.combinations(labelToContigs[label], 2):
			if node1 != node2:
				if (node1, node2) not in graph and (node2, node1) not in graph:
					totalProb = getProb(contigToLabels, contigLabelsToProb, node1, node2)
					if (totalProb > newEdgeProb):
						orgWeight = getMedWeight(orgGraph, node1, node2)
						newWeight = ((1 - alpha) * orgWeight) + (totalProb * alpha)
						ofile.write(node1 + "\t" + node2 + "\t" + str(newWeight) + "\n")
						orgGraph[node1].add((node2, newWeight))
						changesMade += 1
						weightCalc += newWeight
	if (changesMade > 0):
		weightCalc /= changesMade
	return (weightCalc, changesMade)

def run(keys, labelFile, juntoConfigFile, netFile, outdir):
	
	if (netFile == keys["graph_file"]):
		print ("ERROR: junto graph file should be different from the net file otherwise it will be overwritten")
		return 0

	orgGraph = defaultdict(set)
	orgGraphSize = 0
	avgGraphWeight = 0
	with open(netFile, 'r') as ifile:
		for line in ifile:
			edge = (line.strip('\n')).split('\t')
			orgGraph[edge[0]].add((edge[1], float(edge[2])))
			orgGraphSize += 1
			if (float(edge[2]) != 1.1):
				avgGraphWeight += float(edge[2])
	avgGraphWeight /= orgGraphSize

	diff = 10000
	prevDiff = 0
	iters = 1
	i = 0

	with open(keys["graph_file"], 'w') as f:
		for node1 in orgGraph:
			for node2, weight in orgGraph[node1]:
				if node1 == node2:
					f.write(node1 + "\t" + node2 + "\t" + str(0.00001) + "\n")
				else:
					f.write(node1 + "\t" + node2 + "\t" + str(weight) + "\n")

	#while (abs(prevDiff - diff) > (0.1*prevDiff)):
	while (diff > (0.05*prevDiff)):
		i += 1
		print ("Started iteration number: " + str(i) + "\n")

		if (i == 2):
			prevDiff = diff;

		tempfile = open("temp.txt", 'w')
		with open(keys["graph_file"], 'r') as ifile:
			for line in ifile:
				edge = (line.strip('\n')).split('\t')
				node1 = edge[0]
				node2 = edge[1]
				weight = float(edge[2])
				if node1 == node2:
					tempfile.write(node1 + "\t" + node2 + "\t" + str(0.00001) + "\n")
				else:
					tempfile.write(node1 + "\t" + node2 + "\t" + str(weight) + "\n")
			
		tempfile.close()
		call(["mv", "temp.txt", keys["graph_file"]])

		call(["junto", "config", juntoConfigFile])
		call(["cp", keys["output_file"], keys["seed_file"]])

		print("Done running label propogation: processing results.")
		graph = {}
		graphSize = 0
		with open(keys["graph_file"], 'r') as ifile:
			for line in ifile:
				edge = (line.strip('\n')).split('\t')
				if edge[0] != edge[1]:
					graph[(edge[0], edge[1])] = float(edge[2])
				else:
					graph[(edge[0], edge[1])] = 1.1
				graphSize += 1

		labelToContigs, contigToLabels, contigLabelsToProb = readLabels(keys, i)
		print("Done reading labels.")
		sizeNewGraph = 0

		with open(keys["graph_file"], 'w') as ofile:
			#write previous edges with new weights
			(avgOldWeight, temp) = changeEdgeWeights(orgGraph, graph, contigToLabels, contigLabelsToProb, ofile)
			sizeNewGraph += temp
			#add new edges between contigs with same labels
			(avgNewWeight, temp) = addNewEdges(orgGraph, graph, contigToLabels, labelToContigs, contigLabelsToProb, ofile)
			sizeNewGraph += temp

		oldGraphWeight = avgGraphWeight
		avgGraphWeight = (avgOldWeight + avgNewWeight)/2
		
		with open(keys["seed_file"], 'w') as ofile:
			for contig in contigToLabels:
				curProb = 0
				for label in contigToLabels[contig]:
					temp = contigLabelsToProb[(contig, label)]
					ofile.write(contig + "\t" + label + "\t" + str(temp) + "\n")
					curProb += temp
				if (curProb < 1):
					ofile.write(contig + "\t" + "IGNORE" + "\t" + str(1-curProb) + "\n")

		diff = abs(sizeNewGraph - graphSize)
		print("Size diff: {}".format(diff))

		with open(juntoConfigFile, "w") as configFile:
                	for x in keys:
                        	configFile.write(x + " = " + keys[x] + "\n")
                	configFile.write("data_format = edge_factored\n")
                	configFile.write("iters = " + str(iters) + "\n")
               		configFile.write("prune_threshold = 0\n")
                	configFile.write("algo = adsorption\n")



from setuptools import setup

setup(name='biogrouper',
      python_requires="<3.0",
      version='0.1.4',
      scripts=['bin/Grouper'],
      description='Graph-based clustering and annotation for improved de novo transcriptome analysis',
      url='https://github.com/COMBINE-lab/Grouper',
      author='Laraib Malik, Fatemeh Almodaresi, Rob Patro',
      author_email='rob.patro@cs.stonybrook.edu',
      license='BSD with attribution',
      packages=['grouper'],
      install_requires=[
          'PyYAML',
          'coloredlogs',
          'click',
          'networkx==1.11',
          'numpy',
          'pandas',
          'tqdm',
          'statistics'
      ],
      zip_safe=False)
